/* Copyright (c) 2012 Cloudera, Inc. All rights reserved. */

define([
  'knockout',
  'cloudera/common/Humanize',
  'cloudera/common/I18n',
  'cloudera/common/TimeUtil',
  'cloudera/Util'], function(ko, Humanize, I18n, TimeUtil, Util) {
  var commandState = { details: 0, listing: 1, notCopied: 2, failed: 3};

  /**
   * This will add properties to the command needed for displaying the listing and status log data.
   * It needs to be removed before persisting or it will fail to deserialize on the server.
   */
  function prepareCommandForDisplay(command, context) {
    command.loadingListing = ko.observable(false);
    command.listing = ko.observableArray();
    command.loadingNotCopied = ko.observable(false);
    command.notCopied = ko.observableArray();
    var listingLimit = context.options.listingLimit;

    /**
     * This handler function responds to the ko.infinitescroll event fired when the end of the listing containers
     * scrollable bound is reached.
     */
    command.infiniteScrollListing = function(scrollState) {
      var hdfsResult = command.hdfsResult || 
              (command.hiveResult && command.hiveResult.dataReplicationResult);
      // If the command doesn't have and hdfsResult property, the listing log won't have been written.
      if (!hdfsResult) {
        scrollState.complete = true;
        return;
      }
      command.loadingListing(true);
      // Format a URL that includes the current offset and limit.
      $.getJSON(context.options.replicationListingUri
          .replace('{commandId}', command.id())
          .replace('{offset}', scrollState.offset)
          .replace('{limit}', listingLimit), function(data) {
        // Update the listing observable with the value returned from the server.
        command.listing(data);
        command.loadingListing(false);
        // Update the offset with the additional results returned from the server.
        scrollState.offset = scrollState.offset + data.length;
        // If the server returns less results than we requested, the use has reached the end of the results.
        if (data.length < listingLimit) {
          scrollState.complete = true;
        }
      });
    };

    /**
     * This handler function responds to the ko.infinitescroll event fired when the end of the notCopied containers
     * scrollable bound is reached.
     */
    command.infiniteScrollNotCopied = function(scrollState) {
      var hdfsResult = command.hdfsResult || (command.hiveResult && command.hiveResult.dataReplicationResult);
      // If the command doesn't have an hdfsResult property or all files were successfully
      // copied, the status log won't have been written.
      if (!hdfsResult ||
          (hdfsResult.numFilesCopyFailed() === 0 && hdfsResult.numFilesSkipped() === 0)) {
        scrollState.complete = true;
        return;
      }
      command.loadingNotCopied(true);
      //Format a URL that includes the current offset and limit.
      $.getJSON(context.options.replicationStatusUri
          .replace('{commandId}', command.id())
          .replace('{offset}', scrollState.offset)
          .replace('{limit}', listingLimit), function(data) {
        // Update the notCopied observable with the value returned from the server.
        command.notCopied(data);
        command.loadingNotCopied(false);
        // Update the offset with the additional results returned from the server.
        scrollState.offset = scrollState.offset + data.length;
        // If the server returns less results than we requested, the use has reached the end of the results.
        if (data.length < listingLimit) {
          scrollState.complete = true;
        }
      });
    };
  }

  return function(context) {
    var self = this;
    /**
     * If the user lacks permissions required to modify the replication jobs, this can be used to hide
     * elements in the UI.
     */
    self.readOnly = context.options.readOnly;
    
    self.contentLoaded = ko.observable(false);

    /**
     * Identifies whether or not the sources have been loaded.
     */
    self.sourcesLoaded = ko.observable(false);

    /**
     * This observable contains all of the DistCP source services available.
     */
    self.sources = ko.observableArray();
    
    /**
     * This observable is set to the schedule that should be expanded in the UI/
     */
    self.expandedSchedule = ko.observable();

    /**
     * This observable provides a list of all replication schedules.
     */
    self.replicationSchedules = ko.observableArray();
    
    /**
     * This observable is set to the command that should be expanded in the UI.
     */
    self.focusedCommand = ko.observable();
    
    /**
     * This observable contains a valued from {commandState} that identifies the which state the UI should present.
     */
    self.focusedCommandState = ko.observable();
    
    self.commandState = commandState;
    
    self.pathTemplate = context.options.pathTemplate;

    self.scheduleToDelete = ko.observable();

    self.prepareCommandForDisplay =  prepareCommandForDisplay;

    /**
     * A helper function retrieve the HdfsResults for a given command.
     * @param command A command for which to retrieve the HdfsResults.
     * @return object
     */
    self.retrieveHdfsResults = function(command) {
      return command.hdfsResult || (command.hiveResult && command.hiveResult.dataReplicationResult);
    };

    /**
     * A helper function to identify whether or not a listing log is available for a given command.
     * @param command A command for which to evaluate availability of a listing.
     * @return Truthy / Falsy
     */
    self.commandListingAvailable = function(command) {
      return self.retrieveHdfsResults(command);
    };

    /**
     * A helper function to identify whether or not a given command has associated errors.
     * @param command A command for which to evaluate availability of a listing.
     * @return Truthy / Falsy
     */
    self.commandErrorAvailable = function(command) {
      var hdfsResult = self.retrieveHdfsResults(command);
      if (hdfsResult) {
        return hdfsResult.setupError && hdfsResult.setupError();
      } else {
        return command.hiveResult && command.hiveResult.errors().length;
      }
    };

    /**
     * A helper function to identify whether or not a status log is available for a given command.
     * @param command A command for which to evaluate availability of a listing.
     * @return Truthy / Falsy
     */
    self.commandStatusAvailable = function(command) {
      var hdfsResult = self.retrieveHdfsResults(command);
      if (hdfsResult) {
        return (hdfsResult.numFilesCopyFailed && hdfsResult.numFilesCopyFailed()) ||
            (hdfsResult.numFilesSkipped && hdfsResult.numFilesSkipped());
      }
      return false;
    };

    self.setupErrorToLines = function(observable) {
      if (observable) {
        var setupError = ko.utils.unwrapObservable(observable);
        if (setupError) {
          return setupError.split('\n');
        }
      }
    };
    
    self.formatDisplayStatus = function(command) {
      var result = '',
        args = (command.hdfsResult || command.hiveResult),
      // Because in the case of hive the hdfsArguments are moved under the hiveArguments, this is needed.
        hdfsResult = command.hiveResult ? args.dataReplicationResult : args,
        dryRun = args ? ko.utils.unwrapObservable(args.dryRun) : false;

      if (command.active()) {
        if (!args || args.progress() === 0) {
          result = I18n.t('ui.initializing');
        } else if (dryRun) {
          result = args.progress() + '%';
        } else if (hdfsResult) {
          result = I18n.t('ui.replication.ofProcessed',
              hdfsResult.numFilesCopied() +
              hdfsResult.numFilesSkipped() +
              hdfsResult.numFilesCopyFailed(),
              hdfsResult.numFilesExpected());
        }
      } else {
        result = command.endTime ? Humanize.humanizeDateTimeMedium(moment(ko.utils.unwrapObservable(command.endTime)).toDate()) : '';
      }

      if (dryRun) {
        result += I18n.t('ui.replication.dryRunSuffix');
      }
      
      return result;
    };

    /**
     * Formats a display string for the status of a schedule based on the peer's status and the last ran command.
     *
     * @param schedule The schedule to format the message for.
     * @return {string} The formatted string to display for the schedule.
     */
    self.formatScheduleStatus = function(schedule) {
      var lastRun = schedule.lastRun(),
          result;
      if (schedule.invalid()) {
        var args = (schedule.hdfsArguments || schedule.hiveArguments);
        result = I18n.t(args.sourceService.peerName? "ui.invalidPeerSourceService" : "ui.invalidSourceService",
            context.formatReference(args.sourceService));
      // This is a hack: When the schedule is create we immediately request the list of schedules, in some edge cases
      // a schedule that should run for the first time immediately hasn't started. This results in a 15 sec delay before
      // we report initializing.
      } else if ((!lastRun || !lastRun.active()) &&
              schedule.nextRun() &&
              (moment(schedule.nextRun()) - TimeUtil.getServerNow()) < 5000) {
        result = I18n.t("ui.initializing");
      } else if (lastRun) {
        if (lastRun.active()) {
          result = self.formatDisplayStatus(lastRun);
        } else {
          result = self.formatCommandResult(lastRun);
        }
      }
      return result || '';
    };
    
    /**
     * This function will identify the right status message for a given command.
     * command: A command that the display string should be formatted for.
     */
    self.formatCommandResult = function(command) {
      var result;
      if(command.active() || !command.resultMessage) {
        result = '';
      } else if (command.success && command.success()) {
        var hdfsResult = command.hdfsResult || (command.hiveResult && command.hiveResult.dataReplicationResult),
          numFilesExpected = hdfsResult && hdfsResult.numFilesExpected(),
          numFilesCopied = hdfsResult && hdfsResult.numFilesCopied(),
          numFilesDryRun = hdfsResult && hdfsResult.numFilesDryRun();
        if (hdfsResult && hdfsResult.dryRun()) {
          result = I18n.t('ui.replication.availableFor', numFilesDryRun);
        } else if (numFilesCopied !== undefined && numFilesExpected !== undefined) {
          result = I18n.t('ui.replication.ofCopied', numFilesCopied, numFilesExpected);
        } else {
          result = command.resultMessage ? command.resultMessage() : I18n.t('label.resultUnavailable');
        }
      } else {
        result = command.resultMessage();
      }
      return result;
    };

    self.showCommand = function(command) {
      Util.setWindowLocation(context.options.commandDetailsUri.replace('{commandId}', command.id()));
    };

    self.selectSource = function() {
      var schedule = context
        .editViewModel
        .schedule();

      (schedule.hdfsArguments || schedule.hiveArguments).sourceService(this.value);
    };

    /**
     * This function will toggle the paused state of a schedule.
     */
    self.toggleSchedule = function() {
      this.paused(!this.paused());
      context.persistSchedule(context.options, this, true, function() {
        context.loadSchedules();
      });
    };

    /**
     * This function will delete a schedule.
     */
    self.deleteSchedule = function() {
      $.ajax(context.options.replicationUri + encodeURIComponent(self.scheduleToDelete().id()), {
        type : 'DELETE',
        dataType : 'json',
        success : function() {
          context.loadSchedules();
        }
      });
      $('#deleteWarningModal').modal('hide');
    };

    self.initDeleteSchedule = function() {
      self.scheduleToDelete(this);
      $('#deleteWarningModal').modal('show');
    };

    self.editSchedule = function() {
      context.editViewModel.update(true);
      context.editViewModel.schedule(this);
      context.editViewModel.show();
    };

    /**
     * This function will initiate an immediate replication and record the 
     * command to the schedules history.
     */
    self.runReplication = function() {
      context.triggerReplication(this.id());
    };

    /**
     * This function will initiate an immediate replication in "Dry Run" mode 
     * and record the command to the schedules history.
     */
    self.dryRunReplication = function() {
      context.triggerReplication(this.id(), true);
    };

    /**
     * This will display the history of the associated replication schedule, or 
     * if the first command is active, it will show that command.
     */
    self.displayCommandOrHistory = function() {
      var lastCommand = this.history && this.history().length && this.history()[0];
      if (lastCommand.active && lastCommand.active()) {
        self.displayCommand.call(lastCommand);
      } else {
        self.toggleHistory.call(this);
      }
    };
    
    self.toggleHistory = function() {
      if (this.id() !== self.expandedSchedule()) {
        self.expandedSchedule(this.id());
        self.focusedCommand(undefined);
      } else {
        self.expandedSchedule(undefined);
        self.focusedCommand(undefined);
      }
    };

    /**
     * This will display the active command for the associated replication schedule.
     */
    self.displayCommand = function() {
      Util.setWindowLocation(context.options.commandDetailsUri.replace('{commandId}', this.id()));
    };
    
    self.blurCommand = function() {
      self.focusedCommand(undefined);
    };
    
    /**
     * Focuses or blurs a command/state based on the command and state passed.
     * command: The command to focus or blur.
     * state: The {commandState} to focus or blur. 
     */
    self.toggleCommandState = function(command, state) {
      if (command.id() !== (self.focusedCommand() && self.focusedCommand().id())) {
        if (!command.loadingListing) {
          prepareCommandForDisplay(command, context);
        }
        self.focusedCommand(command);
        self.focusedCommandState(state);
      } else if (self.focusedCommandState() !== state) {
        self.focusedCommandState(state);
      } else {
        self.focusedCommand(undefined);
      }
    };
    
    self.toggleCommandDetails = function() {
      self.toggleCommandState(this, commandState.details);
    };
    
    self.toggleCommandListing = function() {
      self.toggleCommandState(this, commandState.listing);
    };
    
    self.toggleCommandNotCopied = function() {
      self.toggleCommandState(this, commandState.notCopied);
    };
    
    self.toggleCommandError = function() {
      self.toggleCommandState(this, commandState.failed);
    };
  };
});
